from pandos.futures.store import FutureStore
from pandos.futures.future import Future

from pandos.settings import get_logger
from pandos.maturity import MaturityLevel


MaturityLevel.PRE_ALPHA.set_module(
    file=__file__,
    logger=get_logger(name=__name__)
)



from dataclasses import dataclass
from functools import wraps
from typing import (
    Callable,
    Generic,
    Generator,
    TypeVar,
)


from pandos.settings import get_logger
from pandos.exceptions import pandos_exceptions
from pandos.maturity import MaturityLevel


MaturityLevel.ALPHA.set_module(
    file=__file__,
    logger=get_logger(name=__name__)
)


A = TypeVar("A")
B = TypeVar("B")


@dataclass(frozen=True, slots=True)
class EitherFunctionWrapper(Generic[A]):
    target: Callable[..., A]

    def either(self, *args, **kwargs) -> 'Either[A]':
        try:
            return Right[A].from_value(x=self.target(*args, **kwargs))
        except Exception as e:
            return Left(exception=e)

    def __call__(self, *args, **kwargs) -> A:
        return self.target(*args, **kwargs)


class Either(Generic[A]):

    @classmethod
    def decorator(cls, function: Callable):
        return wraps(function, updated=())(EitherFunctionWrapper(target=function))

    def __iter__(self):
        match self:
            case right if isinstance(right, Right):
                yield from [right.value]
            case left if isinstance(left, Left):
                yield from []

    @classmethod
    def comprehension(cls, generator: Generator):
        try:
            return Either.from_value(x=next(generator))
        except Exception as e:
            return Left(exception=e)

    @classmethod
    def from_value(cls, x: A, disable_exception_to_left: bool = False) -> 'Either[A]':
        if not disable_exception_to_left and isinstance(x, Exception):
            return Left(exception=x)
        return Right(value=x)

    def flat_map(self, function: Callable[[A], 'Either[B]']) -> 'Either[B]':
        raise NotImplementedError

    def map(self, function: Callable[[A], B]) -> 'Either[B]':
        raise NotImplementedError

    def resolve(self) -> A:
        match self:
            case right if isinstance(right, Right):
                return right.value
            case left if isinstance(left, Left):
                raise left.exception
            case _:
                return pandos_exceptions.MONADS_EITHER_INCONSISTENCY.throw(
                    message="Either Monad inconsistency detected when calling `resolve` method."
                )


@dataclass(frozen=True, slots=True)
class Right(Either[A]):
    value: A

    def flat_map(self, function: Callable[[A], Either[B]]) -> Either[B]:
        # Consider that your right will transform to a left if an exception is encountered
        try:
            return function(self.value)
        except Exception as e:
            return Left(exception=e)

    def map(self, function: Callable[[A], B]) -> Either[B]:
        return self.flat_map(function=lambda value: Right[B].from_value(x=function(value)))


@dataclass(frozen=True, slots=True)
class Left(Either[A]):
    exception: Exception

    def flat_map(self, function: Callable[[A], Either[B]]) -> Either[B]:
        return Left(exception=self.exception)  # Propagate exception

    def map(self, function: Callable[[A], B]) -> Either[B]:
        return Left(exception=self.exception)  # Propagate exception



import os
from setuptools import setup, find_namespace_packages
from typing import (
    List,
    Optional,
)


# Environ variables
PANDOS_CODEBASE_PATH = os.environ.get(
    "PANDOS_CODEBASE_PATH",
    default=os.path.join("src", "main")
)
PANDOS_DISABLE_MYPYC = int(os.environ.get(
    "PANDOS_DISABLE_MYPYC",
    default="0",
))
PANDOS_SETUP_MOCK = int(os.environ.get(
    "PANDOS_SETUP_MOCK",
    default="0",
))


if not PANDOS_DISABLE_MYPYC:
    from mypyc.build import mypycify


with open("README.md", "r") as file:
    readme = file.read()


with open("requirements.txt", "r") as file:
    requirements = [req for req in file.read().splitlines() if req and not req.startswith("#")]


with open(os.path.join(PANDOS_CODEBASE_PATH, "pandos", "version"), "r") as file:
    version = file.read().strip()


with open("LICENSE", "r") as file:
    license_content = file.read()


def get_source_files(
    ignore_mark: Optional[str] = None,
    skip_dirs: Optional[List[str]] = None,
    skip_files: Optional[List[str]] = None,
) -> List[str]:
    skip_dirs = skip_dirs or []
    skip_files = skip_files or []

    def traverse(path: str):
        for _, dirs, files in os.walk(path):
            node_files = [
                os.path.join(path, file)
                for file in files
                if file.endswith(".py") and file not in skip_files
            ]
            return node_files + [
                file
                for directory in dirs
                for file in traverse(os.path.join(path, directory))
                if directory not in skip_dirs
            ]
    return [
        source_file_handler.close() or source_file
        for source_file in traverse(PANDOS_CODEBASE_PATH)
        for source_file_handler in [open(source_file, "r")]
        if ignore_mark is None or ignore_mark not in source_file_handler.read()
    ]


def get_packages(here: str = ".", pandos_prefix: str = "pandos") -> List[str]:
    return [
        package
        for package in find_namespace_packages(where=here)
        if package.startswith(pandos_prefix)
    ]


mypyc_configs = [
    # "--disallow-untyped-defs",  # TODO: Enable this option eventually for better performance
    "--ignore-missing-imports",
]


mypyc_target_files = get_source_files(
    ignore_mark="# pandos-mypyc-ignore-file",
    skip_dirs=[
        "__pycache__",
    ],
    skip_files=[
        "__init__.py"
    ],
)


packages = get_packages(here=PANDOS_CODEBASE_PATH)


if not PANDOS_SETUP_MOCK:
    setup(
        name="pandos",
        version=version,
        description=(
            "The `pandos core` package contains the baseline implementation of the pandos base abstractions."
        ),
        long_description=readme,
        long_description_content_type="text/markdown",
        author="rhdzmota",
        author_email="info@rhdzmota.com",
        url="https://github.com/pando-systems/pandos",
        # https://pypi.org/classifiers/
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",
            "Typing :: Typed",
            "Topic :: Scientific/Engineering",
            "Topic :: System :: Distributed Computing",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.10",
        ],
        package_dir={
            "": PANDOS_CODEBASE_PATH,
        },
        package_data={
            "": [
                "version",
            ]
        },
        include_package_data=True,
        packages=packages,
        scripts=[
            os.path.join("bin", "pandos"),
        ],
        install_requires=requirements,
        python_requires=">=3.10, <4",
        license=license_content,
        zip_safe=False,
        **({} if PANDOS_DISABLE_MYPYC else dict(ext_modules=mypycify(mypyc_configs + mypyc_target_files))),
    )
else:
    # This is temporal, we should eventually remove this!
    import json

    debugging_info = {
        "packages": packages,
        "mypyc_configs": mypyc_configs,
        "mypyc_target_files": mypyc_target_files,
    }

    print(json.dumps(debugging_info, default=str, indent=4))



import os
from typing import (
    List,
)

from dataclasses import dataclass


@dataclass(frozen=True)
class Version:
    name: str
    value: str

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value

    @property
    def components(self) -> List[str]:
        return self.value.split(".")

    @property
    def major(self) -> int:
        component, *_ = self.components
        return int(component)

    @property
    def minor(self) -> int:
        _, component, *_ = self.components
        return int(component)

    @property
    def patch(self) -> int:
        *_, component = self.components
        return int(component)

    @classmethod
    def from_path(cls, dirpath: str, name: str):
        for file in os.listdir(dirpath):
            if file.lower().endswith("version"):
                filepath = os.path.join(dirpath, file)
                break
        else:
            raise ValueError("Version file not found for package name: " + name)

        with open(filepath, "r") as version_file:
            version_value = version_file.readline().strip()  # TODO: Validate version pattern via regex
            return cls(name=name, value=version_value)


version = Version.from_path(name="pandos", dirpath=os.path.dirname(__file__))
